#ifndef WASMTIME_COMPONENT_H
#define WASMTIME_COMPONENT_H

#include <wasmtime/component/component.h>
#include <wasmtime/component/func.h>
#include <wasmtime/component/instance.h>
#include <wasmtime/component/linker.h>
#include <wasmtime/component/val.h>

#endif // WASMTIME_COMPONENT_H
